import numpy as np
from util import MelodistTestCase


class TestPrecipitation(MelodistTestCase):
    def test_equal(self):
        station = self.station
        station.disaggregate_precipitation('equal')

        p = station.data_daily.precip
        pd = station.data_disagg.precip
        pdd = pd.resample('D').sum()

        assert np.allclose(p, pdd, atol=1e-3, equal_nan=True)

    def test_cascade(self):
        station = self.station
        station.statistics.calc_precipitation_stats()
        station.disaggregate_precipitation('cascade')

        p = station.data_daily.precip
        pd = station.data_disagg.precip
        pdd = pd.resample('D').sum()
        pos = p >= 0

        assert np.allclose(station.data_daily.precip[pos], pdd[pos], atol=1e-3, equal_nan=True)



import numpy as np
import util


class TestTemperature(util.MelodistTestCase):
    def test_sine_min_max(self):
        station = self.station
        station.disaggregate_temperature(method='sine_min_max')

        dd = station.data_daily
        pos = util.notna_temp_days(dd, 'minmax')
        td = util.extract_days(station.data_disagg.temp, dd.index[pos])
        tdd = td.resample('D').mean().loc[pos]

        assert np.all(np.isfinite(td))
        assert np.all(tdd >= dd.tmin.min())
        assert np.all(tdd <= dd.tmax.max())
        assert (dd.temp[pos] - tdd).abs().quantile(.99) < 2

    def test_mean_course_min_max(self):
        station = self.station
        station.statistics.calc_temperature_stats()
        station.disaggregate_temperature(method='mean_course_min_max')

        dd = station.data_daily
        pos = util.notna_temp_days(dd, 'minmax')
        td = util.extract_days(station.data_disagg.temp, dd.index[pos])
        tdd = td.resample('D').mean().loc[pos]
        tdd_min = td.resample('D').min().loc[pos]
        tdd_max = td.resample('D').max().loc[pos]

        assert np.all(np.isfinite(td))
        assert np.all(tdd >= dd.tmin.min())
        assert np.all(tdd <= dd.tmax.max())
        assert np.allclose(dd.tmin[pos], tdd_min, atol=1e-3)
        assert np.allclose(dd.tmax[pos], tdd_max, atol=1e-3)

    def test_sine_mean(self):
        station = self.station
        station.disaggregate_temperature(method='sine_mean')

        dd = station.data_daily
        pos = util.notna_temp_days(dd, 'mean')
        td = util.extract_days(station.data_disagg.temp, dd.index[pos])
        tdd = td.resample('D').mean().loc[pos]
        tdd_min = td.resample('D').min().loc[pos]
        tdd_max = td.resample('D').max().loc[pos]

        assert np.all(np.isfinite(td))
        assert np.allclose(dd.temp[pos], tdd, atol=1e-3)
        assert np.allclose(
            (dd.tmax - dd.tmin)[pos],
            tdd_max - tdd_min,
            atol=1e-3
        )

    def test_mean_course_mean(self):
        station = self.station
        station.statistics.calc_temperature_stats()
        station.disaggregate_temperature(method='mean_course_mean')

        dd = station.data_daily
        pos = util.notna_temp_days(dd, 'mean')
        td = util.extract_days(station.data_disagg.temp, dd.index[pos])
        tdd = td.resample('D').mean().loc[pos]
        tdd_min = td.resample('D').min().loc[pos]
        tdd_max = td.resample('D').max().loc[pos]

        assert np.all(np.isfinite(td))
        assert np.allclose(dd.temp[pos], tdd, atol=1e-3)
        assert np.allclose(
            (dd.tmax - dd.tmin)[pos],
            tdd_max - tdd_min,
            atol=1e-3
        )



import copy


# based on http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named
class Bunch(dict):
    def __init__(self, **kw):
        dict.__init__(self, kw)
        self.__dict__ = self

    def __copy__(self):
        copy = type(self)(**self)
        return copy

    def __deepcopy__(self, memo):
        return Bunch(**copy.deepcopy(dict(self)))

    def copy(self):
        return self.__copy__()



import numpy as np
from util import MelodistTestCase


class TestWind(MelodistTestCase):
    def test_generic(self):
        for method in ('equal', 'cosine', 'random'):
            station = self.station
            station.statistics.calc_wind_stats()
            station.disaggregate_wind(method)

            wd = station.data_disagg.wind

            assert wd.notnull().sum() > 0
            assert not np.any(wd < 0)

            if method == 'equal':
                assert np.allclose(
                    station.data_daily.wind,
                    wd.resample('D').mean(),
                    atol=1e-3,
                    equal_nan=True,
                )



import numpy as np
import util


class TestRadiation(util.MelodistTestCase):
    def nighttime_radiation_iszero(self, s, tol=0):
        return np.allclose(s[(s.index.hour >= 22) | (s.index.hour <= 4)], 0, atol=tol)

    def test_pot_rad(self):
        station = self.station
        station.disaggregate_radiation('pot_rad')

        dd = station.data_daily
        pos = dd.glob.notna()
        r = dd.glob[pos]
        rd = util.extract_days(station.data_disagg.glob, dd.index[pos])
        rdd = rd.resample('D').mean().loc[pos]

        assert np.all(np.isfinite(rd))
        assert np.allclose(r, rdd, atol=1e-3, equal_nan=True)
        assert self.nighttime_radiation_iszero(rd)

    def test_pot_rad_via_bc(self):
        station = self.station
        station.disaggregate_radiation('pot_rad_via_bc')

        dd = station.data_daily
        pos = util.notna_temp_days(dd, 'minmax')
        rd = util.extract_days(station.data_disagg.glob, dd.index[pos])

        assert np.all(np.isfinite(rd))
        assert self.nighttime_radiation_iszero(rd)

    def test_mean_course(self):
        station = self.station
        station.statistics.calc_radiation_stats()
        station.disaggregate_radiation('mean_course')

        dd = station.data_daily
        pos = dd.glob.notna()
        r = dd.glob[pos]
        rd = util.extract_days(station.data_disagg.glob, dd.index[pos])

        assert np.all(np.isfinite(rd))
        assert self.nighttime_radiation_iszero(rd, tol=10)
        assert np.allclose(r.resample('M').mean(), rd.resample('M').mean(), rtol=0.1)



################################################################################
# This file is part of MELODIST - MEteoroLOgical observation time series       #
# DISaggregation Tool.                                                         #
#                                                                              #
# Copyright (C) 2016-2023 Florian Hanzer, Kristian Förster, Benjamin Winter,   #
# Thomas Marke                                                                 #
#                                                                              #
# MELODIST is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU General Public License as published by the Free         #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# MELODIST is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more        #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU General Public License along with #
# this program.  If not, see <http://www.gnu.org/licenses/>.                   #
################################################################################
import numpy as np
import pandas as pd

import melodist
import melodist.util.util as util


def disaggregate_humidity(
    data_daily,
    method='equal',
    temp=None,
    a0=None,
    a1=None,
    kr=None,
    month_hour_precip_mean=None,
    preserve_daily_mean=False,
):
    """general function for humidity disaggregation

    Args:
        daily_data: daily values
        method: keyword specifying the disaggregation method to be used
        temp: hourly temperature time series (necessary for some methods)
        kr: parameter for linear_dewpoint_variation method (6 or 12)
        month_hour_precip_mean: [month, hour, precip(y/n)] categorical mean values
        preserve_daily_mean: if True, correct the daily mean values of the disaggregated
            data with the observed daily means.

    Returns:
        Disaggregated hourly values of relative humidity.
    """
    assert method in (
        'equal',
        'minimal',
        'dewpoint_regression',
        'min_max',
        'linear_dewpoint_variation',
        'month_hour_precip_mean',
    ), 'Invalid option'

    if method == 'equal':
        hum_disagg = melodist.distribute_equally(data_daily.hum)
    elif method in ('minimal', 'dewpoint_regression', 'linear_dewpoint_variation'):
        if method == 'minimal':
            a0 = 0
            a1 = 1

        assert a0 is not None and a1 is not None, 'a0 and a1 must be specified'
        tdew_daily = a0 + a1 * data_daily.tmin

        tdew = melodist.distribute_equally(tdew_daily)

        if method == 'linear_dewpoint_variation':
            assert kr is not None, 'kr must be specified'
            assert kr in (6, 12), 'kr must be 6 or 12'
            tdew_delta = 0.5 * np.sin(
                (temp.index.hour + 1) * np.pi / kr - 3.0 * np.pi / 4.0
            )  # eq. (21) from Debele et al. (2007)

            tdew_nextday = tdew.shift(-24)
            tdew_nextday.iloc[-24:] = tdew.iloc[-24:]  # copy the last day

            # eq. (20) from Debele et al. (2007):
            # (corrected - the equation is wrong both in Debele et al. (2007) and Bregaglio et al.
            # (2010) - it should be (T_dp,day)_(d+1) - (T_dp,day)_d instead of the other way around)
            tdew += temp.index.hour / 24.0 * (tdew_nextday - tdew) + tdew_delta

        sat_vap_press_tdew = util.vapor_pressure(tdew, 100)
        sat_vap_press_t = util.vapor_pressure(temp, 100)
        hum_disagg = pd.Series(index=temp.index, data=100 * sat_vap_press_tdew / sat_vap_press_t)
    elif method == 'min_max':
        assert (
            'hum_min' in data_daily.columns and 'hum_max' in data_daily.columns
        ), 'Minimum and maximum humidity must be present in data frame'

        hmin = melodist.distribute_equally(data_daily.hum_min)
        hmax = melodist.distribute_equally(data_daily.hum_max)
        tmin = melodist.distribute_equally(data_daily.tmin)
        tmax = melodist.distribute_equally(data_daily.tmax)

        hum_disagg = hmax + (temp - tmin) / (tmax - tmin) * (hmin - hmax)
    elif method == 'month_hour_precip_mean':
        assert month_hour_precip_mean is not None

        precip_equal = melodist.distribute_equally(
            data_daily.precip
        )  # daily precipitation equally distributed to hourly values
        hum_disagg = pd.Series(index=precip_equal.index, dtype=float)
        locs = list(zip(hum_disagg.index.month, hum_disagg.index.hour, precip_equal > 0))
        hum_disagg[:] = month_hour_precip_mean.loc[locs].values

    if preserve_daily_mean:
        daily_mean_df = pd.DataFrame(
            data=dict(obs=data_daily.hum, disagg=hum_disagg.resample('D').mean())
        )
        bias = melodist.util.distribute_equally(daily_mean_df.disagg - daily_mean_df.obs)
        bias = bias.fillna(0)
        hum_disagg -= bias

    return hum_disagg.clip(0, 100)


def calculate_dewpoint_regression(hourly_data_obs, return_stats=False):
    temphum = hourly_data_obs[['temp', 'hum']]

    tdew = melodist.util.dewpoint_temperature(temphum.temp, temphum.hum).resample('D').mean()
    tmin = temphum.temp.groupby(temphum.index.normalize()).min()
    df = pd.DataFrame(data=dict(tmin=tmin, tdew=tdew)).dropna(how='any')

    return util.linregress(df.tmin, df.tdew, return_stats=return_stats)


def calculate_month_hour_precip_mean(hourly_data_obs):
    daily_precip_yesno = hourly_data_obs.precip.resample('D').sum() > 0
    daily_precip_yesno.index.name = None
    hum = hourly_data_obs.hum
    hum.index.name = None
    wet = daily_precip_yesno.loc[hum.index.date].values
    mhp_mean = hum.groupby([hum.index.month, hum.index.hour, wet]).mean()

    return mhp_mean



import numpy as np
from util import MelodistTestCase


class TestHumidity(MelodistTestCase):
    def test_generic(self):
        for method in ('equal', 'minimal', 'dewpoint_regression',
                       'linear_dewpoint_variation', 'min_max', 'month_hour_precip_mean'):
            station = self.station
            station.statistics.calc_humidity_stats()
            station.disaggregate_temperature()
            station.disaggregate_humidity(method)

            hd = station.data_disagg.hum

            assert hd.notnull().sum() > 0
            assert not np.any(hd <= 0)
            assert not np.any(hd > 100)

    def test_equal(self):
        station = self.station
        station.disaggregate_humidity('equal')

        hd = station.data_disagg.hum
        hdd = hd.resample('D').mean()

        assert np.allclose(
            hdd,
            station.statistics.data.hum.resample('D').mean(),
            atol=1e-3,
            equal_nan=True,
        )



################################################################################
# This file is part of MELODIST - MEteoroLOgical observation time series       #
# DISaggregation Tool.                                                         #
#                                                                              #
# Copyright (C) 2016-2023 Florian Hanzer, Kristian Förster, Benjamin Winter,   #
# Thomas Marke                                                                 #
#                                                                              #
# MELODIST is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU General Public License as published by the Free         #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# MELODIST is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more        #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU General Public License along with #
# this program.  If not, see <http://www.gnu.org/licenses/>.                   #
################################################################################
import numpy as np
import pandas as pd
import scipy.stats


def hourly_index(daily_index, fill_gaps=False):
    index = pd.date_range(start=daily_index.min(), end=daily_index.max().replace(hour=23), freq='H')

    # remove days that are not in the daily index:
    if not fill_gaps and len(index) > len(daily_index) * 24:
        df = pd.DataFrame(index=index.date, data=dict(hour=index.hour, keep=True))
        df.index = df.index.to_datetime()
        dropdates = list(set(index.date) - set(daily_index.date))
        df.loc[dropdates, 'keep'] = False
        df = df[df.keep]
        index = pd.DatetimeIndex(
            [
                pd.datetime(y, m, d, h)
                for y, m, d, h in zip(df.index.year, df.index.month, df.index.day, df.hour)
            ]
        )

    return index


def distribute_equally(daily_data, divide=False):
    """Obtains hourly values by equally distributing the daily values.

    Args:
        daily_data: daily values
        divide: if True, divide resulting values by the number of hours in
            order to preserve the daily sum (required e.g. for precipitation).

    Returns:
        Equally distributed hourly values.
    """

    index = hourly_index(daily_data.index)
    hourly_data = daily_data.reindex(index)
    hourly_data = hourly_data.groupby(hourly_data.index.day).transform(
        lambda x: x.fillna(method='ffill', limit=23)
    )

    if divide:
        hourly_data /= 24

    return hourly_data


def vapor_pressure(temp, hum):
    """
    Calculates vapor pressure from temperature and humidity after Sonntag (1990).

    Args:
        temp: temperature values
        hum: humidity value(s). Can be scalar (e.g. for calculating saturation vapor pressure).

    Returns:
        Vapor pressure in hPa.
    """

    if np.isscalar(hum):
        hum = np.zeros(temp.shape) + hum

    assert temp.shape == hum.shape

    positives = np.array(temp >= 273.15)
    vap_press = np.zeros(temp.shape) * np.nan
    vap_press[positives] = (
        6.112
        * np.exp((17.62 * (temp[positives] - 273.15)) / (243.12 + (temp[positives] - 273.15)))
        * hum[positives]
        / 100.0
    )
    vap_press[~positives] = (
        6.112
        * np.exp((22.46 * (temp[~positives] - 273.15)) / (272.62 + (temp[~positives] - 273.15)))
        * hum[~positives]
        / 100.0
    )

    return vap_press


def dewpoint_temperature(temp, hum):
    """computes the dewpoint temperature

    Parameters
    ----
    temp :      temperature [K]
    hum :       relative humidity


    Returns
        dewpoint temperature in K
    """
    assert temp.shape == hum.shape

    vap_press = vapor_pressure(temp, hum)

    positives = np.array(temp >= 273.15)
    dewpoint_temp = temp.copy() * np.nan
    dewpoint_temp[positives] = (
        243.12
        * np.log(vap_press[positives] / 6.112)
        / (17.62 - np.log(vap_press[positives] / 6.112))
    )
    dewpoint_temp[~positives] = (
        272.62
        * np.log(vap_press[~positives] / 6.112)
        / (22.46 - np.log(vap_press[~positives] / 6.112))
    )

    return dewpoint_temp + 273.15


def linregress(x, y, return_stats=False):
    """linear regression calculation

    Parameters
    ----
    x :         independent variable (series)
    y :         dependent variable (series)
    return_stats : returns statistical values as well if required (bool)


    Returns
    ----
    list of parameters (and statistics)
    """
    a1, a0, r_value, p_value, stderr = scipy.stats.linregress(x, y)

    retval = a1, a0
    if return_stats:
        retval += r_value, p_value, stderr

    return retval


def get_sun_times(dates, lon, lat, time_zone):
    """Computes the times of sunrise, solar noon, and sunset for each day.

    Parameters
    ----
    dates:      datetime
    lat :       latitude in DecDeg
    lon :       longitude in DecDeg
    time_zone : timezone


    Returns
    ----
    DataFrame:  [sunrise, sunnoon, sunset, day length] in dec hours
    """

    df = pd.DataFrame(index=dates, columns=['sunrise', 'sunnoon', 'sunset', 'daylength'])

    doy = np.array([(d - d.replace(day=1, month=1)).days + 1 for d in df.index])  # day of year

    # Day angle and declination after Bourges (1985):
    day_angle_b = np.deg2rad((360.0 / 365.25) * (doy - 79.346))

    declination = np.deg2rad(
        0.3723
        + 23.2567 * np.sin(day_angle_b)
        - 0.7580 * np.cos(day_angle_b)
        + 0.1149 * np.sin(2 * day_angle_b)
        + 0.3656 * np.cos(2 * day_angle_b)
        - 0.1712 * np.sin(3 * day_angle_b)
        + 0.0201 * np.cos(3 * day_angle_b)
    )

    # Equation of time with day angle after Spencer (1971):
    day_angle_s = 2 * np.pi * (doy - 1) / 365.0
    eq_time = (
        12.0
        / np.pi
        * (
            0.000075
            + 0.001868 * np.cos(day_angle_s)
            - 0.032077 * np.sin(day_angle_s)
            - 0.014615 * np.cos(2 * day_angle_s)
            - 0.040849 * np.sin(2 * day_angle_s)
        )
    )

    #
    standard_meridian = time_zone * 15.0
    delta_lat_time = (lon - standard_meridian) * 24.0 / 360.0

    omega_nul_arg = -np.tan(np.deg2rad(lat)) * np.tan(declination)
    omega_nul = np.arccos(omega_nul_arg)
    sunrise = 12.0 * (1.0 - (omega_nul) / np.pi) - delta_lat_time - eq_time
    sunset = 12.0 * (1.0 + (omega_nul) / np.pi) - delta_lat_time - eq_time

    # as an approximation, solar noon is independent of the below mentioned
    # cases:
    sunnoon = 12.0 * (1.0) - delta_lat_time - eq_time

    # $kf 2015-11-13: special case midnight sun and polar night
    # CASE 1: MIDNIGHT SUN
    # set sunrise and sunset to values that would yield the maximum day
    # length even though this a crude assumption
    pos = omega_nul_arg < -1
    sunrise[pos] = sunnoon[pos] - 12
    sunset[pos] = sunnoon[pos] + 12

    # CASE 2: POLAR NIGHT
    # set sunrise and sunset to values that would yield the minmum day
    # length even though this a crude assumption
    pos = omega_nul_arg > 1
    sunrise[pos] = sunnoon[pos]
    sunset[pos] = sunnoon[pos]

    daylength = sunset - sunrise

    # adjust if required
    sunrise[sunrise < 0] += 24
    sunset[sunset > 24] -= 24

    df.sunrise = sunrise
    df.sunnoon = sunnoon
    df.sunset = sunset
    df.daylength = daylength

    return df


def detect_gaps(dataframe, timestep, print_all=False, print_max=5, verbose=True):
    """checks if a given dataframe contains gaps and returns the number of gaps

    This function checks if a dataframe contains any gaps for a given temporal
    resolution that needs to be specified in seconds. The number of gaps
    detected in the dataframe is returned.

    Args:
        dataframe: A pandas dataframe object with index defined as datetime
        timestep (int): The temporal resolution of the time series in seconds
            (e.g., 86400 for daily values)
        print_all (bool, opt): Lists every gap on the screen
        print_mx (int, opt): The maximum number of gaps listed on the screen in
            order to avoid a decrease in performance if numerous gaps occur
        verbose (bool, opt): Enables/disables output to the screen

    Returns:
        The number of gaps as integer. Negative values indicate errors.
    """
    gcount = 0
    msg_counter = 0
    warning_printed = False
    try:
        n = len(dataframe.index)
    except:
        print('Error: Invalid dataframe.')
        return -1
    for i in range(0, n):
        if i > 0:
            time_diff = dataframe.index[i] - dataframe.index[i - 1]
            if time_diff.delta / 1e9 != timestep:
                gcount += 1
                if print_all or (msg_counter <= print_max - 1):
                    if verbose:
                        print(
                            'Warning: Gap in time series found between %s and %s'
                            % (dataframe.index[i - 1], dataframe.index[i])
                        )
                    msg_counter += 1
                if msg_counter == print_max and verbose and not warning_printed:
                    print(
                        'Warning: Only the first %i gaps have been listed. Try to increase '
                        'print_max parameter to show more details.'
                        % msg_counter
                    )
                    warning_printed = True
    if verbose:
        print('%i gaps found in total.' % (gcount))
    return gcount


def drop_incomplete_days(dataframe, shift=0):
    """truncates a given dataframe to full days only

    This function truncates a given pandas dataframe (time series) to full days
    only, thus dropping leading and tailing hours of incomplete days. Please
    note that this methodology only applies to hourly time series.

    Args:
        dataframe: A pandas dataframe object with index defined as datetime
        shift (unsigned int, opt): First hour of daily recordings. For daily
            recordings of precipitation gages, 8 would be the first hour of
            the subsequent day of recordings since daily totals are
            usually recorded at 7. Omit defining this parameter if you intend
            to pertain recordings to 0-23h.

    Returns:
        A dataframe with full days only.
    """
    dropped = 0
    if shift > 23 or shift < 0:
        print("Invalid shift parameter setting! Using defaults.")
        shift = 0
    first = shift
    last = first - 1
    if last < 0:
        last += 24
    try:
        # todo: move this checks to a separate function
        n = len(dataframe.index)
    except:
        print('Error: Invalid dataframe.')
        return dataframe

    delete = list()

    # drop heading lines if required
    for i in range(0, n):
        if dataframe.index.hour[i] == first and dataframe.index.minute[i] == 0:
            break
        else:
            delete.append(i)
            dropped += 1

    # drop tailing lines if required
    for i in range(n - 1, 0, -1):
        if dataframe.index.hour[i] == last and dataframe.index.minute[i] == 0:
            break
        else:
            delete.append(i)
            dropped += 1
    # print("The following rows have been dropped (%i in total):" % dropped)
    # print(delete)
    return dataframe.drop(index=dataframe.index[delete])


def prepare_interpolation_data(data_daily, column_hours):
    start_date = data_daily.index[0]
    end_date = data_daily.index[-1]

    data = pd.Series()

    for column, hour in column_hours.items():
        index = pd.date_range(
            start=start_date.replace(hour=hour), end=end_date.replace(hour=hour), freq='D'
        )
        s = pd.Series(index=index, data=data_daily[column].values)
        data = data.append(s)

    data = data.sort_index()
    data = data.reindex(hourly_index(data_daily.index))

    return data


def daily_from_hourly(df):
    """Aggregates data (hourly to daily values) according to the characteristics
    of each variable (e.g., average for temperature, sum for precipitation)

    Args:
        df: dataframe including time series with one hour time steps

    Returns:
        dataframe (daily)
    """

    df_daily = pd.DataFrame()

    if 'temp' in df:
        df_daily['temp'] = df.temp.resample('D').mean()
        df_daily['tmin'] = df.temp.groupby(df.temp.index.date).min()
        df_daily['tmax'] = df.temp.groupby(df.temp.index.date).max()

    if 'precip' in df:
        df_daily['precip'] = df.precip.resample('D').sum()

    if 'glob' in df:
        df_daily['glob'] = df.glob.resample('D').mean()

    if 'hum' in df:
        df_daily['hum'] = df.hum.resample('D').mean()

    if 'hum' in df:
        df_daily['hum_min'] = df.hum.groupby(df.hum.index.date).min()

    if 'hum' in df:
        df_daily['hum_max'] = df.hum.groupby(df.hum.index.date).max()

    if 'wind' in df:
        df_daily['wind'] = df.wind.resample('D').mean()

    if 'ssd' in df:
        df_daily['ssd'] = df.ssd.resample('D').sum() / 60  # minutes to hours

    df_daily.index.name = None
    return df_daily


def calculate_mean_daily_course_by_month(data_hourly, normalize=False):
    data_hourly = data_hourly.copy()
    data_hourly.index.name = None

    df = data_hourly.groupby([data_hourly.index.month, data_hourly.index.hour]).mean()
    df = df.reset_index().pivot(columns='level_0', index='level_1')
    df.columns = df.columns.droplevel()  # remove MultiIndex
    df.columns.name = None
    df.index.name = None

    if normalize:
        df = (df - df.min()) / (df.max() - df.min())  # normalize values to 0-1 range

    return df



import copy

import numpy as np
from util import MelodistTestCase, setup_station


class TestStation(MelodistTestCase):
    def setUp(self):
        self._station = setup_station()

    @property
    def station(self):
        return copy.copy(self._station)

    def test_copy(self):
        s1 = self.station
        s2 = copy.deepcopy(s1)

        s1.statistics.temp.max_delta = 3
        s1.data_disagg.temp[:] = 273.15

        assert s1.statistics.temp.max_delta != s2.statistics.temp.max_delta
        assert not np.allclose(s1.data_disagg, s2.data_disagg, equal_nan=True)



import tempfile

from numpy.testing import assert_equal
from pandas.testing import assert_frame_equal, assert_series_equal
from util import MelodistTestCase

import melodist


class TestIO(MelodistTestCase):
    def test_json(self):
            ss = self.station.statistics
            ss.calc_temperature_stats()
            ss.calc_precipitation_stats()
            ss.calc_humidity_stats()
            ss.calc_radiation_stats()
            ss.calc_wind_stats()

            with tempfile.NamedTemporaryFile() as tmp:
                ss.to_json(tmp.name)
                tmp.seek(0)
                ss2 = melodist.StationStatistics.from_json(tmp.name)

            assert_series_equal_kwargs = dict(
                check_dtype=False,
                check_index_type=False,
            )
            assert_frame_equal_kwargs = dict(
                check_index_type=False,
                check_column_type=False,
            )

            assert_series_equal(
                ss.temp.max_delta,
                ss2.temp.max_delta,
                **assert_series_equal_kwargs,
            )
            assert_frame_equal(
                ss.temp.mean_course,
                ss2.temp.mean_course,
                **assert_frame_equal_kwargs,
            )

            assert_equal(ss.precip.months, ss2.precip.months)
            assert all([cs1 == cs2 for cs1, cs2 in zip(ss.precip.stats, ss2.precip.stats)])

            assert ss.hum.a0 == ss2.hum.a0
            assert ss.hum.a1 == ss2.hum.a1
            assert ss.hum.kr == ss2.hum.kr
            assert_series_equal(
                ss.hum.month_hour_precip_mean,
                ss2.hum.month_hour_precip_mean,
                **assert_series_equal_kwargs,
            )

            assert_frame_equal(ss.glob.angstroem, ss2.glob.angstroem, **assert_frame_equal_kwargs)
            assert_frame_equal(ss.glob.bristcamp, ss2.glob.bristcamp, **assert_frame_equal_kwargs)
            assert_frame_equal(
                ss.glob.mean_course,
                ss2.glob.mean_course,
                **assert_frame_equal_kwargs,
            )

            assert ss.wind.a == ss2.wind.a
            assert ss.wind.b == ss2.wind.b
            assert ss.wind.t_shift == ss2.wind.t_shift



################################################################################
# This file is part of MELODIST - MEteoroLOgical observation time series       #
# DISaggregation Tool.                                                         #
#                                                                              #
# Copyright (C) 2016-2023 Florian Hanzer, Kristian Förster, Benjamin Winter,   #
# Thomas Marke                                                                 #
#                                                                              #
# MELODIST is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU General Public License as published by the Free         #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# MELODIST is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more        #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU General Public License along with #
# this program.  If not, see <http://www.gnu.org/licenses/>.                   #
################################################################################
from . __version__ import __version__

from . import cascade
from .data_io import *
from .humidity import *
from .precipitation import *
from .radiation import *
from .station import Station
from .stationstatistics import StationStatistics
from .temperature import *
from .util.util import distribute_equally
from .wind import *

columns_daily = [
    'tmean',
    'tmin',
    'tmax',
    'precip',
    'glob',
    'ssd',
    'hum',
    'wind',
]

columns_hourly = [
    'temp',
    'precip',
    'glob',
    'hum',
    'wind',
]


class Options:  # (deprecated)
    TEMP_SINE_CURVE = 'sine'
    TEMP_LINEAR = 'linear'
    EQUAL_DISTRIBUTION = 'equal'
    PREC_CASCADE = 'cascade'
    PREC_MASTER_STATION = 'masterstation'
    HUMIDITY_MINIMAL = 'minimal'
    HUMIDITY_DEWPOINT_REGRESSION = 'dewpoint_regression'
    HUMIDITY_MIN_MAX = 'min_max'
    HUMIDITY_LINEAR_DEWPOINT_VARIATION = 'linear_dewpoint_variation'
    WIND_FITTED_COSINE = 'cosine'
    WIND_RANDOM_EXPONENTIAL = 'random'
    POT_RAD = 'pot_rad'
    POT_RAD_VIA_SSD = 'pot_rad_via_ssd'
    POT_RAD_VIA_BC = 'pot_rad_via_bc'



import unittest

import numpy as np
import pandas as pd

import melodist


class Test5MinutePrecipitation(unittest.TestCase):
    def test_cascade(self):
        data = pd.read_csv(
            'examples/testdata_precip5min.csv.gz',
            names=['time', 'precip'],
            parse_dates=True,
            index_col=0,
        )
        precip_daily = data.precip.resample('D').sum()

        cascopt = melodist.build_casc(data, hourly=False, level=9, percentile=90)
        precip_disagg = melodist.disagg_prec_cascade(
            precip_daily,
            cascopt[0],
            hourly=False,
            level=9,
        )

        assert np.allclose(
            precip_daily,
            precip_disagg.resample('D').sum(),
            atol=1e-3,
            equal_nan=True,
        )



################################################################################
# This file is part of MELODIST - MEteoroLOgical observation time series       #
# DISaggregation Tool.                                                         #
#                                                                              #
# Copyright (C) 2016-2023 Florian Hanzer, Kristian Förster, Benjamin Winter,   #
# Thomas Marke                                                                 #
#                                                                              #
# MELODIST is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU General Public License as published by the Free         #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# MELODIST is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more        #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU General Public License along with #
# this program.  If not, see <http://www.gnu.org/licenses/>.                   #
################################################################################
import numpy as np
import pandas as pd
import scipy.optimize

import melodist


def _cosine_function(x, a, b, t_shift):
    """Generate a diurnal course of windspeed according to the cosine function

    Args:
        x: series of euqally distributed windspeed values
        a: parameter a for the cosine function
        b: parameter b for the cosine function
        t_shift: parameter t_shift for the cosine function

    Returns:
        series including diurnal course of windspeed.
    """

    mean_wind, t = x
    return a * mean_wind * np.cos(np.pi * (t - t_shift) / 12) + b * mean_wind


def disaggregate_wind(wind_daily, method='equal', a=None, b=None, t_shift=None):
    """General function for windspeed disaggregation

    Args:
        wind_daily: daily values
        method: keyword specifying the disaggregation method to be used
        a: parameter a for the cosine function
        b: parameter b for the cosine function
        t_shift: parameter t_shift for the cosine function

    Returns:
        Disaggregated hourly values of windspeed.
    """
    assert method in ('equal', 'cosine', 'random'), 'Invalid method'

    wind_eq = melodist.distribute_equally(wind_daily)

    if method == 'equal':
        wind_disagg = wind_eq
    elif method == 'cosine':
        assert None not in (a, b, t_shift)
        wind_disagg = _cosine_function(
            np.array([wind_eq.values, wind_eq.index.hour]), a, b, t_shift
        )
    elif method == 'random':
        wind_disagg = wind_eq * (-np.log(np.random.rand(len(wind_eq)))) ** 0.3

    return wind_disagg


def fit_cosine_function(wind):
    """Fit a cosine function to observed hourly windspeed data

    Args:
        wind: observed hourly windspeed data

    Returns:
        parameters needed to generate diurnal features of windspeed using a cosine function
    """
    wind_daily = wind.groupby(wind.index.date).mean()
    wind_daily_hourly = pd.Series(
        index=wind.index, data=wind_daily.loc[wind.index.date].values
    )  # daily values evenly distributed over the hours

    df = pd.DataFrame(data=dict(daily=wind_daily_hourly, hourly=wind)).dropna(how='any')
    x = np.array([df.daily, df.index.hour])
    popt, pcov = scipy.optimize.curve_fit(_cosine_function, x, df.hourly)

    return popt
